//-----------------------------------------------------------------------------
//           Name: ogl_point_sprites.cpp
//         Author: Kevin Harris
//  Last Modified: 09/09/02
//    Description: This sample demonstrates how to use hardware accelerated 
//                 point sprites with OpenGL.
//-----------------------------------------------------------------------------

#define STRICT

#include <windows.h>
#include <mmsystem.h>
#include <gl\gl.h>
#include <gl\glaux.h>
#include "resource.h"
#include "vector3f.h"
#include <stdio.h>

#include "Common.h"
//-----------------------------------------------------------------------------
// FUNCTION POINTERS FOR OPENGL EXTENSIONS
//-----------------------------------------------------------------------------
#include "glext.h"

PFNGLPOINTPARAMETERFEXTPROC  glPointParameterfEXT  = NULL;
PFNGLPOINTPARAMETERFVEXTPROC glPointParameterfvEXT = NULL;

PFNGLPOINTPARAMETERINVPROC   glPointParameteriNV   = NULL;
PFNGLPOINTPARAMETERIVNVPROC  glPointParameterivNV  = NULL;

//-----------------------------------------------------------------------------
// GLOBALS
//-----------------------------------------------------------------------------
HWND   g_hWnd      = NULL;
HDC	   g_hDC       = NULL;
HGLRC  g_hRC       = NULL;
GLuint g_textureID = -1;

float g_fSpinX = 0.0f;
float g_fSpinY = 0.0f;



//-----------------------------------------------------------------------------
// PROTOTYPES
//-----------------------------------------------------------------------------
int WINAPI WinMain(HINSTANCE hInstance,HINSTANCE hPrevInstance, 
				   LPSTR lpCmdLine, int nCmdShow);
LRESULT CALLBACK WindowProc(HWND hWnd, UINT msg, WPARAM wParam, LPARAM lParam);
void init(void);
void render(void);
void shutDown(void);

void initPointSprites(void);
void updatePointSprites(void);
void renderPointSprites(void);

float getRandomMinMax(float fMin, float fMax);
vector3f getRandomVector(void);

#define OGL_PI    ((FLOAT)  3.141592654f)

////////////////////////////////////
///////////*********////////////////
////////////////////////////////////
#define IX(i,j,k) ( (i)+(Nx+2)*(j)+(Nx+2)*(Ny+2)*(k) )
/* external definitions (from solver.c) */

//extern void dens_step ( int N, float * x, float * x0, float * u, float * v, float *w, float diff, float dt );
//extern void vel_step ( int N, float * u, float * v, float *w, float * u0, float * v0, float *w0, float visc, float dt );

extern void dens_step ( int Nx, int Ny, int Nz, float * x, float * x0, float * u, float * v, float * w, float diff, float dt );
extern void vel_step ( int Nx, int Ny, int Nz, float * u, float * v, float * w, float * u0, float * v0, float * w0, float visc, float dt );
/* global variables */

static int Nx, Ny, Nz;
static float dt, diff, visc;
static float force, source;
static int dvel;

static float * u, * v, * w, * u_prev, * v_prev, *w_prev;
static float * dens, * dens_prev;

static int win_id;
static int win_x, win_y;
static int mouse_down[3];
static int omx, omy, omz, mx, my, mz;

static float rx=0, ry=0, rz=0;

//////////////////////////////////
static void get_from_UI ( float * d, float * u, float * v )
{
	int i, j, size = (Nx+2)*(Ny+2)*(Nz+2);

	for ( i=0 ; i<size ; i++ ) {
		u[i] = v[i] = d[i] =w_prev[i]= 0.0f;
	}

	if ( !mouse_down[0] && !mouse_down[2] ) return;

	i = (int)((       mx /(float)win_x)*Nx+1);
	j = (int)(((win_y-my)/(float)win_y)*Ny+1);

	if ( i<1 || i>Nx || j<1 || j>Ny ) return;


	if ( mouse_down[0] ) {
	  u[IX(i,j, 1)] = force * (mx-omx);
	  v[IX(i,j, 1)] = force * (omy-my);
	}

	if ( mouse_down[2] ) {
		d[IX(i,j, 1)] = source;
	}

	omx = mx;
	omy = my;

	return;
}
static void free_data ( void )
{
	if ( u ) free ( u );
	if ( v ) free ( v );
	if ( w ) free ( w );
	if ( u_prev ) free ( u_prev );
	if ( v_prev ) free ( v_prev );
	if ( w_prev ) free ( w_prev );
	if ( dens ) free ( dens );
	if ( dens_prev ) free ( dens_prev );
}

static void clear_data ( void )
{
	int i, size=(Nx+2)*(Ny+2)*(Nz+2);

	for ( i=0 ; i<size ; i++ ) {
		u[i] = v[i] = w[i]=u_prev[i] = v_prev[i] =w_prev[i]= dens[i] = dens_prev[i] = 0.0f;
	}
}

static int allocate_data ( void )
{
	int size = (Nx+2)*(Ny+2)*(Nz+2);

	u			= (float *) malloc ( size*sizeof(float) );
	v			= (float *) malloc ( size*sizeof(float) );
	w			= (float *) malloc ( size*sizeof(float) );
	u_prev		= (float *) malloc ( size*sizeof(float) );
	v_prev		= (float *) malloc ( size*sizeof(float) );
	w_prev		= (float *) malloc ( size*sizeof(float) );

	dens		= (float *) malloc ( size*sizeof(float) );	
	dens_prev	= (float *) malloc ( size*sizeof(float) );

	if ( !u || !v || !w || !u_prev || !v_prev || !w_prev || !dens || !dens_prev ) {
		return ( 0 );
	}

	return ( 1 );
}


//////////*************/////////

//-----------------------------------------------------------------------------
// Name: WinMain()
// Desc: The application's entry point
//-----------------------------------------------------------------------------
int WINAPI WinMain(	HINSTANCE hInstance,
					HINSTANCE hPrevInstance,
					LPSTR     lpCmdLine,
					int       nCmdShow )
{
	WNDCLASSEX winClass; 
	MSG        uMsg;

    memset(&uMsg,0,sizeof(uMsg));

	winClass.lpszClassName = "MY_WINDOWS_CLASS";
	winClass.cbSize        = sizeof(WNDCLASSEX);
	winClass.style         = CS_HREDRAW | CS_VREDRAW | CS_OWNDC;
	winClass.lpfnWndProc   = WindowProc;
	winClass.hInstance     = hInstance;
    winClass.hIcon	       = LoadIcon(hInstance, (LPCTSTR)IDI_OPENGL_ICON);
    winClass.hIconSm	   = LoadIcon(hInstance, (LPCTSTR)IDI_OPENGL_ICON);
	winClass.hCursor       = LoadCursor(NULL, IDC_ARROW);
	winClass.hbrBackground = (HBRUSH)GetStockObject(BLACK_BRUSH);
	winClass.lpszMenuName  = NULL;
	winClass.cbClsExtra    = 0;
	winClass.cbWndExtra    = 0;
	
	if(!RegisterClassEx(&winClass) )
		return E_FAIL;

	g_hWnd = CreateWindowEx( NULL, "MY_WINDOWS_CLASS", "OpenGL Point Sprites",
						     WS_OVERLAPPEDWINDOW | WS_VISIBLE,
					         0, 0, 640, 480, NULL, NULL, hInstance, NULL );

	if( g_hWnd == NULL )
		return E_FAIL;

    ShowWindow( g_hWnd, nCmdShow );
    UpdateWindow( g_hWnd );

	init();
    initPointSprites();

	IVector vec(1,1,1);
	Grid<IVector> gr(IVector(10,10,10));

	int k=gr.getCell(IVector(0,0,0)).x;

	while( uMsg.message != WM_QUIT )
	{
		if( PeekMessage( &uMsg, NULL, 0, 0, PM_REMOVE ) )
		{
			TranslateMessage( &uMsg );
			DispatchMessage( &uMsg );
		}
        else
			render();
	}

	shutDown();

    UnregisterClass( "MY_WINDOWS_CLASS", hInstance );

	return uMsg.wParam;
}

//-----------------------------------------------------------------------------
// Name: WindowProc()
// Desc: The window's message handler
//-----------------------------------------------------------------------------
LRESULT CALLBACK WindowProc( HWND   hWnd, 
							 UINT   msg, 
							 WPARAM wParam, 
							 LPARAM lParam )
{
    static POINT ptLastMousePosit;
	static POINT ptCurrentMousePosit;
	static bool bMousing;

    switch( msg )
	{
        case WM_KEYDOWN:
		{
			switch( wParam )
			{
				case VK_ESCAPE:
					PostQuitMessage(0);
				break;
			}
		}
        break;

		case WM_LBUTTONDOWN:
		{
			ptLastMousePosit.x = ptCurrentMousePosit.x = LOWORD (lParam);
            ptLastMousePosit.y = ptCurrentMousePosit.y = HIWORD (lParam);
			bMousing = true;
		}
		break;

		case WM_LBUTTONUP:
		{
			bMousing = false;

		}
		break;


        case WM_MBUTTONDOWN:
		{

			//////cc//////
			omx = mx = LOWORD (lParam);
			omx = my = HIWORD (lParam);
			mouse_down[0] = 1;
			//////cc/////
		}
		break;

		case WM_MBUTTONUP:
		{

			//////cc///////
			omx = mx = LOWORD (lParam);
			omx = my = HIWORD (lParam);
			mouse_down[0] = 0;
		}
		break;
		
		//////cc//////////
		case WM_RBUTTONDOWN:
		{

			//////cc//////
			omx = mx = LOWORD (lParam);
			omx = my = HIWORD (lParam);
			mouse_down[2] = 1;
			//////cc/////
		}
		break;
		case WM_RBUTTONUP:
		{
			bMousing = false;

			//////cc///////
			omx = mx = LOWORD (lParam);
			omx = my = HIWORD (lParam);
			mouse_down[2] = 0;
		}
		break;
		/////////cc/////////

		case WM_MOUSEMOVE:
		{
			ptCurrentMousePosit.x = LOWORD (lParam);
			ptCurrentMousePosit.y = HIWORD (lParam);

			if( bMousing )
			{
				g_fSpinX -= (ptCurrentMousePosit.x - ptLastMousePosit.x);
				g_fSpinY -= (ptCurrentMousePosit.y - ptLastMousePosit.y);
			}
			
			ptLastMousePosit.x = ptCurrentMousePosit.x;
            ptLastMousePosit.y = ptCurrentMousePosit.y;

			//////////cc/////////
			mx = LOWORD (lParam);
			my = HIWORD (lParam);

		}
		break;

		case WM_CLOSE:
		{
			PostQuitMessage(0);	
		}

        case WM_DESTROY:
		{
            PostQuitMessage(0);
		}
        break;
		
		default:
		{
			return DefWindowProc( hWnd, msg, wParam, lParam );
		}
		break;
	}

	return 0;
}

//-----------------------------------------------------------------------------
// Name: init()
// Desc: 
//-----------------------------------------------------------------------------
void init( void )
{
	GLuint PixelFormat;

	PIXELFORMATDESCRIPTOR pfd;
	memset(&pfd, 0, sizeof(PIXELFORMATDESCRIPTOR));

    pfd.nSize      = sizeof(PIXELFORMATDESCRIPTOR);
    pfd.nVersion   = 1;
    pfd.dwFlags    = PFD_DRAW_TO_WINDOW |PFD_SUPPORT_OPENGL | PFD_DOUBLEBUFFER;
    pfd.iPixelType = PFD_TYPE_RGBA;
    pfd.cColorBits = 32;
    pfd.cDepthBits = 32;
	
	g_hDC = GetDC( g_hWnd );
	PixelFormat = ChoosePixelFormat( g_hDC, &pfd );
	SetPixelFormat( g_hDC, PixelFormat, &pfd);
	g_hRC = wglCreateContext( g_hDC );
	wglMakeCurrent( g_hDC, g_hRC );

	glClearColor( 0.00f, 0.20f, 0.52f, 1.0f );
	glClearDepth(1.0f);
	glEnable(GL_TEXTURE_2D);

	//glEnable(GL_DEPTH_TEST);
	
	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();
	gluPerspective( 45.0f, 640.0f / 480.0f, 0.1f, 100.0f);

	//
	// If the required extensions are present, get the addresses of thier 
	// functions that we wish to use...
	//

	char *ext = (char*)glGetString( GL_EXTENSIONS );

	if( strstr( ext, "GL_EXT_point_parameters" ) == NULL )
	{
		MessageBox(NULL,"GL_EXT_point_parameters extension was not found",
			"ERROR",MB_OK|MB_ICONEXCLAMATION);
		return;
	}
	else
	{
        glPointParameterfEXT  = (PFNGLPOINTPARAMETERFEXTPROC)wglGetProcAddress("glPointParameterfEXT");
        glPointParameterfvEXT = (PFNGLPOINTPARAMETERFVEXTPROC)wglGetProcAddress("glPointParameterfvEXT");

		if( !glPointParameterfEXT || !glPointParameterfvEXT )
		{
			MessageBox(NULL,"One or more GL_EXT_point_parameters functions were not found",
				"ERROR",MB_OK|MB_ICONEXCLAMATION);
			return;
		}
	}

    if( strstr( ext, "GL_NV_point_sprite" ) == NULL )
	{
		MessageBox(NULL,"GL_NV_point_sprite extension was not found",
			"ERROR",MB_OK|MB_ICONEXCLAMATION);
		return;
	}
	else
	{
        glPointParameteriNV   = (PFNGLPOINTPARAMETERINVPROC)wglGetProcAddress("glPointParameteriNV");
        glPointParameterivNV  = (PFNGLPOINTPARAMETERIVNVPROC)wglGetProcAddress("glPointParameterivNV");

		if( !glPointParameteriNV  || !glPointParameterivNV )
		{
			MessageBox(NULL,"One or more GL_NV_point_sprite functions were not found",
				"ERROR",MB_OK|MB_ICONEXCLAMATION);
			return;
		}
	}
    
}

//-----------------------------------------------------------------------------
// Name: shutDown()
// Desc: 
//-----------------------------------------------------------------------------
void shutDown( void )	
{
	if( g_hRC != NULL )
	{
		wglMakeCurrent( NULL, NULL );
		wglDeleteContext( g_hRC );
		g_hRC = NULL;
	}

	if( g_hDC != NULL )
	{
		ReleaseDC( g_hWnd, g_hDC );
		g_hDC = NULL;
	}
}

//-----------------------------------------------------------------------------
// Name: initPointSprites()
// Desc: 
//-----------------------------------------------------------------------------
void initPointSprites( void )
{
	//
	// Load up the point sprite's texture...
	//
	int i, j;

	AUX_RGBImageRec *pTextureImage = auxDIBImageLoad( ".\\particle.bmp" );

    if( pTextureImage != NULL )
	{
        glGenTextures( 1, &g_textureID );

		glBindTexture(GL_TEXTURE_2D, g_textureID);


		int t=pTextureImage->sizeX;
		int size=t*t*4;
		char *tmp=(char*)malloc(size);

		for(i=0; i<t; i++)
			for(j=0; j<t; j++){
				tmp[0+4*i+(4*t)*j]=pTextureImage->data[0+3*i+(3*t)*j];
				tmp[1+4*i+(4*t)*j]=pTextureImage->data[1+3*i+(3*t)*j];
				tmp[2+4*i+(4*t)*j]=pTextureImage->data[2+3*i+(3*t)*j];
				tmp[3+4*i+(4*t)*j]=pTextureImage->data[0+3*i+(3*t)*j]*0.5;
			}
				

		glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MIN_FILTER, GL_LINEAR);
		glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MAG_FILTER, GL_LINEAR);

		glTexImage2D(GL_TEXTURE_2D, 0, GL_RGBA, pTextureImage->sizeX, pTextureImage->sizeY, 0,
				     GL_RGBA, GL_UNSIGNED_BYTE, tmp);
	}

	if( pTextureImage )
	{
		if( pTextureImage->data )
			free( pTextureImage->data );

		free( pTextureImage );
	}

	//
    // If you want to know the max size that a point sprite can be set 
    // to, do this.
	//

    // Query for the max point size supported by the hardware
    float maxSize = 0.0f;
    glGetFloatv( GL_POINT_SIZE_MAX_EXT, &maxSize );
    glPointSize( maxSize );



	/////////
	Nx = 16;
	Ny=16;
	Nz=16;
	dt = 0.40f;
	diff = 0.0f;
	visc = 0.0001f;
	force = 1.0f;
	source = 20.0f;

	if ( !allocate_data () ) exit ( 1 );
	clear_data ();

	win_x = 512;
	win_y = 512;

	/////////

}


//-----------------------------------------------------------------------------
// Name: renderPointSprites()
// Desc: 
//-----------------------------------------------------------------------------
void renderPointSprites( void )
{
    //
    // Set up for blending...
    //

	glEnable( GL_BLEND );
	glBlendFunc( GL_SRC_ALPHA , GL_DST_ALPHA); //GL_ONE );GL_SRC_ALPHA GL_DST_ALPHA
	//glBlendFunc( GL_SRC_ALPHA , GL_ONE_MINUS_SRC_ALPHA);
	//glEnable(GL_COLOR_MATERIAL);

    //
	// Set up the OpenGL state machine for using point sprites...
	//

    // This is how will our point sprite's size will be modified by 
    // distance from the viewer
    float quadratic[] =  { 1.0f, 0.0f, 0.001f };
    glPointParameterfvEXT( GL_DISTANCE_ATTENUATION_EXT, quadratic );

    // Query for the max point size supported by the hardware
    float maxSize = 0.0f;
    glGetFloatv( GL_POINT_SIZE_MAX_EXT, &maxSize );
    glPointSize( maxSize );

    // The alpha of a point is calculated to allow the fading of points 
    // instead of shrinking them past a defined threshold size. The threshold 
    // is defined by GL_POINT_FADE_THRESHOLD_SIZE_EXT and is not clamped to 
    // the minimum and maximum point sizes.
    glPointParameterfEXT( GL_POINT_FADE_THRESHOLD_SIZE_EXT, 60.0f );

    glPointParameterfEXT( GL_POINT_SIZE_MIN_EXT, 40.0f );
    glPointParameterfEXT( GL_POINT_SIZE_MAX_EXT, 45.0 );

    // Specify point sprite texture coordinate replacement mode for each texture unit
    glTexEnvf( GL_POINT_SPRITE_NV, GL_COORD_REPLACE_NV, GL_TRUE );

    // Set the point sprite R coordinate mode
   // glPointParameteriNV( GL_POINT_SPRITE_R_MODE_NV, GL_ZERO );
   // glPointParameteriNV( GL_POINT_SPRITE_R_MODE_NV, GL_S );
   // glPointParameteriNV( GL_POINT_SPRITE_R_MODE_NV, GL_R );

    //
	// Render point sprites...
	//

    glEnable( GL_POINT_SPRITE_NV );

	glBegin( GL_POINTS );
	////////////////
	int i, j, k;
	float x, y, z, hx, hy, hz, d00, d01, d10, d11;
	hx = 3.0f/Nx;
	hy = 3.0f/Ny;
	hz = 3.0f/Nz;

	for(k=1; k<=Nz; k++){
		z = ((k)-0.5f)*hz;
		for ( i=1 ; i<=Nx ; i++ ) {
			x = (i-0.5f)*hx;
			for ( j=1 ; j<=Ny ; j++ ) {
				y = (j-0.5f)*hy;

				d00 = dens[IX(i,j, k)];
				
			
				glColor4f ( 0.7, 0.7, 1.0,d00*d00*10);

				glVertex3f( x-1.5, y-1.5, z-1.5);


			}
		}
	}

	glEnd();
}

//-----------------------------------------------------------------------------
// Name: render()
// Desc: 
//-----------------------------------------------------------------------------
void render( void )
{
    glClear( GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT );
	
	glMatrixMode( GL_MODELVIEW );
    glLoadIdentity();
	glTranslatef( 0.0f, 0.0f, -5.0f );
    glRotatef( -g_fSpinY, 1.0f, 0.0f, 0.0f );
    glRotatef( -g_fSpinX, 0.0f, 1.0f, 0.0f );

	//////////
	get_from_UI ( dens_prev, u_prev, v_prev );
	vel_step ( Nx, Ny, Nz, u, v, w, u_prev, v_prev, w_prev, visc, dt );
	dens_step ( Nx, Ny, Nz, dens, dens_prev, u, v, w, diff, dt );
	//////////

    
	glBindTexture( GL_TEXTURE_2D, g_textureID );

    renderPointSprites();

	SwapBuffers( g_hDC );
}
